from __future__ import annotations

import typing

from django.utils.translation import gettext_lazy as _
from rest_framework.pagination import BasePagination
from rest_framework.utils.urls import replace_query_param
from rest_framework.exceptions import NotFound

from restdoctor.constants import DEFAULT_MAX_PAGE_SIZE, DEFAULT_PAGE_SIZE
from restdoctor.rest_framework.pagination.mixins import SerializerClassPaginationMixin
from restdoctor.rest_framework.pagination.serializers import (
    PageNumberRequestSerializer, PageNumberResponseSerializer, PageNumberUncountedResponseSerializer,
)
from restdoctor.rest_framework.response import ResponseWithMeta

if typing.TYPE_CHECKING:
    from django.db.models import QuerySet
    from rest_framework.request import Request
    from rest_framework.views import APIView

    from restdoctor.rest_framework.pagination.custom_types import OptionalList


class PageNumberPagination(SerializerClassPaginationMixin, BasePagination):
    use_count = True
    page_query_param = 'page'
    page_size_query_param = 'per_page'
    serializer_class_map = {
        'default': PageNumberRequestSerializer,
        'pagination': {
            'response': PageNumberResponseSerializer,
        },
    }

    max_page_size = DEFAULT_MAX_PAGE_SIZE
    default_page_size = DEFAULT_PAGE_SIZE

    invalid_page_message = _('Invalid page.')

    def paginate_queryset(
        self, queryset: QuerySet, request: Request, view: APIView = None,
    ) -> OptionalList:
        serializer_class = self.get_request_serializer_class()
        serializer = serializer_class(data=request.query_params, max_per_page=self.max_page_size)
        serializer.is_valid(raise_exception=False)

        self.per_page = serializer.validated_data.get(self.page_size_query_param, self.default_page_size)
        self.page = serializer.validated_data.get(self.page_query_param, 1)

        self.has_next = False
        self.has_prev = (self.page > 1)

        base_url = request.build_absolute_uri()
        self.base_url = replace_query_param(base_url, self.page_size_query_param, self.per_page)
        self.request = request

        start_offset = (self.page - 1) * self.per_page
        stop_offset = start_offset + self.per_page + 1

        if self.use_count:
            self.total = len(queryset) if isinstance(queryset, list) else queryset.count()
            if self.total and self.per_page:
                self.pages, rem = divmod(self.total, self.per_page)
                if rem:
                    self.pages += 1
            else:
                self.pages = 1
            if self.page > self.pages:
                msg = self.invalid_page_message.format(page_number=self.page)
                raise NotFound(msg)

        paginated = list(queryset[start_offset:stop_offset])

        if len(paginated) > self.per_page:
            self.has_next = True
            del paginated[-1]

        return paginated

    def get_page_link_tmpl(self) -> str:
        url_tmpl = self.request.build_absolute_uri()
        url_tmpl = replace_query_param(url_tmpl, self.page_size_query_param, self.per_page)
        return url_tmpl

    def get_page_link(self, page: typing.Any = 1) -> typing.Optional[str]:
        if page:
            return replace_query_param(self.base_url, self.page_query_param, page)

    def get_paginated_response(self, data: typing.Sequence[typing.Any]) -> ResponseWithMeta:
        meta = {
            self.page_query_param: self.page,
            self.page_size_query_param: self.per_page,
            'url': self.get_page_link(page=self.page),
        }
        if self.use_count:
            meta['total'] = self.total

        if self.use_count and self.pages > 1:
            meta['last_url'] = self.get_page_link(page=self.pages)
        meta['next_url'] = self.get_page_link(page=self.page + 1) if self.has_next else None
        meta['prev_url'] = self.get_page_link(page=self.page - 1) if self.has_prev else None

        return ResponseWithMeta(data=data, meta=meta)


class PageNumberUncountedPagination(PageNumberPagination):
    use_count = False

    serializer_class_map = {
        'default': PageNumberRequestSerializer,
        'pagination': {
            'response': PageNumberUncountedResponseSerializer,
        },
    }



import pytest
from django.core.exceptions import ImproperlyConfigured
from rest_framework.fields import Field

from restdoctor.rest_framework.schema import RestDoctorSchema
from tests.stubs.serializers import MyModelWithoutHelpTextsSerializer


@pytest.mark.parametrize(
    'field,expected_help_text',
    (
        (Field(), None),
        (Field(help_text='help_text'), 'help_text'),
    ),
)
def test_get_field_description_not_raises_success_case(field, expected_help_text):
    schema = RestDoctorSchema()

    result = schema.get_field_description(field)

    assert result == expected_help_text


def test_get_field_description_strict_fail_case(settings):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    schema = RestDoctorSchema()

    with pytest.raises(ImproperlyConfigured):
        schema.get_field_description(Field())


def test_get_field_description_success_case(settings):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    schema = RestDoctorSchema()

    result = schema.get_field_description(Field(help_text='help_text'))

    assert result == 'help_text'


def test_get_field_description_model_field_not_raises_success_case():
    schema = RestDoctorSchema()

    result = schema.get_field_description(
        MyModelWithoutHelpTextsSerializer().fields['abstract_field'])

    assert result is None


def test_get_field_description_model_field_strict_fail_case(settings):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    schema = RestDoctorSchema()

    with pytest.raises(ImproperlyConfigured):
        schema.get_field_description(MyModelWithoutHelpTextsSerializer().fields['abstract_field'])


def test_get_field_description_model_field_success_case(settings):
    settings.API_STRICT_SCHEMA_VALIDATION = True
    schema = RestDoctorSchema()

    result = schema.get_field_description(MyModelWithoutHelpTextsSerializer().fields['timestamp'])

    assert result == 'Event timestamp'



from __future__ import annotations

import contextlib
import typing
import uuid

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.pagination import BasePagination
from rest_framework.utils.urls import replace_query_param, remove_query_param

from restdoctor.constants import DEFAULT_MAX_PAGE_SIZE, DEFAULT_PAGE_SIZE
from restdoctor.rest_framework.pagination.mixins import SerializerClassPaginationMixin
from restdoctor.rest_framework.pagination.serializers import (
    CursorUUIDRequestSerializer, CursorUUIDResponseSerializer, CursorUUIDUncountedResponseSerializer,
)
from restdoctor.rest_framework.response import ResponseWithMeta

if typing.TYPE_CHECKING:
    from django.db.models import QuerySet
    from rest_framework.request import Request
    from rest_framework.views import APIView

    from restdoctor.rest_framework.pagination.custom_types import OptionalList, OptionalLookup


def get_order(default_order: str, serializer_keys: typing.Sequence[str]) -> str:
    if 'after' in serializer_keys:
        return 'after'
    if 'before' in serializer_keys:
        return 'before'
    return default_order


def get_cursor(
    queryset: QuerySet,
    after_uuid: uuid.UUID = None,
    before_uuid: uuid.UUID = None,
    default_order: str = None,
) -> typing.Tuple[typing.Any, str]:
    cursor_obj = None
    if after_uuid is not None:
        with contextlib.suppress(ObjectDoesNotExist):
            cursor_obj = queryset.get(uuid=after_uuid)
            order = 'after'
    elif before_uuid is not None:
        with contextlib.suppress(ObjectDoesNotExist):
            cursor_obj = queryset.get(uuid=before_uuid)
            order = 'before'
    if cursor_obj is None:
        order = default_order or ''
    return cursor_obj, order


class CursorUUIDPagination(SerializerClassPaginationMixin, BasePagination):
    use_count = True

    after_query_param = 'after'
    before_query_param = 'before'
    page_size_query_param = 'per_page'

    lookup_by_field = 'timestamp'
    order_by_field = 'timestamp'
    default_order = 'before'

    serializer_class_map = {
        'default': CursorUUIDRequestSerializer,
        'pagination': {
            'response': CursorUUIDResponseSerializer,
        },
    }

    max_page_size = DEFAULT_MAX_PAGE_SIZE
    default_page_size = DEFAULT_PAGE_SIZE

    def get_lookup(self) -> OptionalLookup:
        if self.cursor_obj:
            lookup_operator = 'gt' if self.order == 'after' else 'lt'
            lookup_keyword = f'{self.lookup_by_field}__{lookup_operator}'
            return {lookup_keyword: getattr(self.cursor_obj, self.lookup_by_field)}

    def paginate_queryset(
        self, queryset: QuerySet, request: Request, view: APIView = None,
    ) -> OptionalList:
        serializer_class = self.get_request_serializer_class()
        serializer = serializer_class(data=request.query_params, max_per_page=self.max_page_size)
        serializer.is_valid(raise_exception=True)

        self.per_page = serializer.validated_data.get(self.page_size_query_param, self.default_page_size)

        after_uuid = serializer.validated_data.get(self.after_query_param)
        before_uuid = serializer.validated_data.get(self.before_query_param)

        order = get_order(self.default_order, serializer.validated_data.keys())
        self.cursor_obj, self.order = get_cursor(queryset, after_uuid, before_uuid, order)

        lookup = self.get_lookup()
        if lookup:
            queryset = queryset.filter(**lookup)

        order_sign = '-' if self.order == 'before' else ''
        order_keyword = f'{order_sign}{self.order_by_field}'
        queryset = queryset.order_by(order_keyword)

        base_url = request.build_absolute_uri()
        self.base_url = replace_query_param(base_url, self.page_size_query_param, self.per_page)
        self.request = request
        self.has_next = False

        start_offset = 0
        stop_offset = start_offset + self.per_page + 1

        if self.use_count:
            self.total = queryset.count()

        paginated = list(queryset[:stop_offset])

        if len(paginated) > self.per_page:
            self.has_next = True
            del paginated[-1]

        if paginated:
            self.page_boundaries = (
                paginated[0].uuid,
                paginated[-1].uuid,
            )
        elif self.cursor_obj:
            self.page_boundaries = (self.cursor_obj.uuid, self.cursor_obj.uuid)
        else:
            self.page_boundaries = (None, None)

        return paginated

    def get_page_link_tmpl(self) -> str:
        url_tmpl = self.request.build_absolute_uri()
        url_tmpl = replace_query_param(url_tmpl, self.page_size_query_param, self.per_page)
        return url_tmpl

    def get_page_link(self, after: typing.Any = None, before: typing.Any = None) -> typing.Optional[str]:
        if after is not None:
            base_url = remove_query_param(self.base_url, self.before_query_param)
            return replace_query_param(base_url, self.after_query_param, after)
        if before is not None:
            base_url = remove_query_param(self.base_url, self.after_query_param)
            return replace_query_param(base_url, self.before_query_param, before)

    def get_paginated_response(self, data: typing.Sequence[typing.Any]) -> ResponseWithMeta:
        meta = {
            self.page_size_query_param: self.per_page,
            'has_next': self.has_next,
        }
        cursor_uuid = None
        if self.cursor_obj:
            cursor_uuid = self.cursor_obj.uuid
        meta[self.order] = cursor_uuid
        meta['url'] = self.get_page_link(**{self.order: cursor_uuid or ''})

        if self.use_count:
            meta['total'] = self.total

        if self.page_boundaries[0]:
            after_idx, before_idx = 0, 1
            if self.order == 'after':
                after_idx, before_idx = 1, 0

            meta['after_url'] = self.get_page_link(
                **{self.after_query_param: self.page_boundaries[after_idx]})
            meta['before_url'] = self.get_page_link(
                **{self.before_query_param: self.page_boundaries[before_idx]})

        return ResponseWithMeta(data=data, meta=meta)


class CursorUUIDUncountedPagination(CursorUUIDPagination):
    use_count = False

    serializer_class_map = {
        'default': CursorUUIDRequestSerializer,
        'pagination': {
            'response': CursorUUIDUncountedResponseSerializer,
        },
    }



import pytest

from restdoctor.rest_framework.schema import RefsSchemaGenerator
from tests.test_unit.test_schema.stubs import SerializerClassMapViewSet, SerializerClassMapView


@pytest.mark.parametrize(
    'path,method,expected_operation_id,expected_responses_required',
    (
        ('/test/', 'GET', 'listLists', ['list_field']),
        ('/test/', 'POST', 'createCreate', ['create_field']),
        ('/test/1234/', 'GET', 'retrieveDefault', ['default_field']),
    ),
)
def test_serializer_class_map_viewset_schema(
    path, method, expected_operation_id, expected_responses_required,
    get_create_view_func,
):
    create_view = get_create_view_func('test', SerializerClassMapViewSet, 'test')

    view = create_view(path, method)
    schema = view.schema
    schema.generator = None
    result = view.schema.get_operation(path, method)
    code, _ = view.schema._get_action_code_description(path, method)
    responses_schema = result['responses'][code]['content']['application/vnd.vendor']['schema']
    responses_properties = responses_schema['properties']['data']
    if 'items' in responses_properties:
        responses_properties = responses_properties['items']

    assert result['operationId'] == expected_operation_id
    assert responses_properties['required'] == expected_responses_required


@pytest.mark.parametrize(
    'path,method,expected_operation_id,expected_responses_required',
    (
        ('/test/', 'GET', 'retrieveList', ['list_field']),
        ('/test/', 'POST', 'createCreate', ['create_field']),
    ),
)
def test_serializer_class_map_view_schema(
    path, method, expected_operation_id, expected_responses_required,
):
    generator = RefsSchemaGenerator()

    view = generator.create_view(SerializerClassMapView.as_view(), method, None)
    schema = view.schema
    schema.generator = None
    result = view.schema.get_operation(path, method)
    code, _ = view.schema._get_action_code_description(path, method)
    responses_schema = result['responses'][code]['content']['application/vnd.vendor']['schema']
    responses_properties = responses_schema['properties']['data']
    if 'items' in responses_properties:
        responses_properties = responses_properties['items']

    assert result['operationId'] == expected_operation_id
    assert responses_properties['required'] == expected_responses_required



import pytest

from django.utils.translation import gettext_lazy as _

from restdoctor.constants import DEFAULT_MAX_PAGE_SIZE, DEFAULT_PAGE_SIZE
from restdoctor.rest_framework.pagination import CursorUUIDPagination
from restdoctor.rest_framework.pagination.page_number import PageNumberPagination
from restdoctor.rest_framework.schema import RestDoctorSchema


@pytest.mark.parametrize(
    ('pagination_class', 'expected_schema'), [
        (
            PageNumberPagination,
            {
                'type': 'object',
                'properties': {
                    'page': {
                        'type': 'integer',
                        'minimum': 1,
                        'description': _('Selected page')
                    },
                    'per_page': {
                        'type': 'integer',
                        'maximum': DEFAULT_MAX_PAGE_SIZE,
                        'description': _('Page size'),
                    },
                    'total': {
                        'type': 'integer',
                        'description': _('Total result size'),
                    },
                    'has_next': {
                        'type': 'boolean',
                        'description': _('Has result next page')
                    },
                    'url': {
                        'type': 'string',
                        'description': _('Current page URL'),
                    },
                    'last_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('Last page URL'),
                    },
                    'next_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('Next page URL'),
                    },
                    'prev_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('Previous page URL'),
                    },
                },
                'required': [
                    'page', 'per_page', 'has_next', 'url',
                    'next_url', 'prev_url', 'total', 'last_url',
                ],
            },
        ),
        (
            CursorUUIDPagination,
            {
                'type': 'object',
                'properties': {
                    'per_page': {
                        'type': 'integer',
                        'maximum': DEFAULT_MAX_PAGE_SIZE,
                        'description': _('Page size'),
                    },
                    'has_next': {
                        'type': 'boolean',
                        'description': _('Has result next page')
                    },
                    'before': {
                        'type': 'string',
                        'format': 'uuid',
                        'nullable': True,
                        'description': _('Before UUID'),
                    },
                    'after': {
                        'type': 'string',
                        'format': 'uuid',
                        'nullable': True,
                        'description': _('After UUID'),
                    },
                    'total': {
                        'type': 'integer',
                        'description': _('Total result size'),
                    },
                    'url': {
                        'type': 'string',
                        'description': _('Current page URL'),
                    },
                    'last_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('Last page URL'),
                    },
                    'after_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('URL of page after cursor position'),
                    },
                    'before_url': {
                        'type': 'string',
                        'nullable': True,
                        'description': _('URL of page before cursor position'),
                    },
                },
                'required': [
                    'per_page', 'has_next', 'url',
                    'after_url', 'before_url', 'total', 'last_url',
                ],
            },
        ),
    ]
)
def test_pagination_schema(pagination_class, expected_schema):
    paginator = pagination_class(view_schema=RestDoctorSchema())
    base_schema = {'properties': {}}

    schema = paginator.get_paginated_response_schema(base_schema)

    assert schema['properties']['meta'] == expected_schema



from __future__ import annotations
from typing import TYPE_CHECKING

from rest_framework.parsers import JSONParser

if TYPE_CHECKING:
    from restdoctor.utils.custom_types import GenericContext


class BestDoctorParser(JSONParser):
    def __init__(self, media_type: str, api_params: GenericContext) -> None:
        self.media_type = media_type
        self.api_params = api_params



import pytest
from django.core.validators import URLValidator
from rest_framework.fields import (
    Field, ReadOnlyField, IntegerField, CharField, EmailField,
    RegexField, URLField, UUIDField,
)

from restdoctor.rest_framework.schema import RestDoctorSchema


url_pattern = URLValidator.regex.pattern.replace('\\Z', '\\z')

@pytest.mark.parametrize(
    'field,expected_schema',
    (
        (Field(), {'type': 'string'}),
        (ReadOnlyField(), {'readOnly': True, 'type': 'string'}),
        (Field(write_only=True), {'writeOnly': True, 'type': 'string'}),
        (Field(allow_null=True), {'nullable': True, 'type': 'string'}),
        (Field(default='default'), {'default': 'default', 'type': 'string'}),
        (Field(help_text='help_text'), {'description': 'help_text', 'type': 'string'}),
    ),
)
def test_basic_fields_schema(field, expected_schema):
    schema = RestDoctorSchema()

    result = schema._get_field_schema(field)

    assert result == expected_schema


@pytest.mark.parametrize(
    'field,expected_schema',
    (
        (IntegerField(min_value=100, max_value=200), {'type': 'integer', 'minimum': 100, 'maximum': 200}),
        (IntegerField(max_value=2147483648), {'type': 'integer', 'maximum': 2147483648, 'format': 'int64'}),
        (CharField(min_length=5, max_length=10), {'type': 'string', 'minLength': 5, 'maxLength': 10}),
        (EmailField(), {'type': 'string', 'format': 'email'}),
        (RegexField('pattern'), {'type': 'string', 'pattern': 'pattern'}),
        (URLField(), {'type': 'string', 'format': 'uri', 'pattern': url_pattern}),
        (UUIDField(), {'type': 'string', 'format': 'uuid'}),
    ),
)
def test_fields_validators_schema(field, expected_schema):
    schema = RestDoctorSchema()

    result = schema._get_field_schema(field)

    assert result == expected_schema



from __future__ import annotations
import typing

from django.utils.translation import gettext_lazy as _
from rest_framework.fields import IntegerField, UUIDField, BooleanField, CharField
from rest_framework.serializers import Serializer

from restdoctor.constants import DEFAULT_PAGE_SIZE, DEFAULT_MAX_PAGE_SIZE


class PerPageSerializerBase(Serializer):
    per_page = IntegerField(default=DEFAULT_PAGE_SIZE, allow_null=True,
                            help_text='Количество элементов на странице')

    def __init__(
        self, max_per_page: int = 200,
        *args: typing.Any, **kwargs: typing.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.max_per_page = max_per_page

    def validate_per_page(self, value: typing.Optional[int]) -> int:
        if value is None:
            return DEFAULT_PAGE_SIZE
        if value > self.max_per_page:
            return self.max_per_page
        return value


class PageNumberRequestSerializer(PerPageSerializerBase):
    page = IntegerField(min_value=1, default=1, help_text='Выбранная страница, начиная с 1')


class PageNumberUncountedResponseSerializer(Serializer):
    page = IntegerField(min_value=1, help_text=_('Selected page'))
    per_page = IntegerField(
        required=True, max_value=DEFAULT_MAX_PAGE_SIZE,
        help_text=_('Page size'),
    )
    has_next = BooleanField(help_text=_('Has result next page'))
    url = CharField(help_text=_('Current page URL'))
    next_url = CharField(help_text=_('Next page URL'), allow_null=True)
    prev_url = CharField(help_text=_('Previous page URL'), allow_null=True)


class PageNumberResponseSerializer(PageNumberUncountedResponseSerializer):
    total = IntegerField(help_text=_('Total result size'))
    last_url = CharField(help_text=_('Last page URL'), allow_null=True)


class CursorUUIDRequestSerializer(PerPageSerializerBase):
    after = UUIDField(required=False, allow_null=True, help_text=_('After UUID'))
    before = UUIDField(required=False, allow_null=True, help_text=_('Before UUID'))


class CursorUUIDUncountedResponseSerializer(Serializer):
    after = UUIDField(required=False, allow_null=True, help_text=_('After UUID'))
    before = UUIDField(required=False, allow_null=True, help_text=_('Before UUID'))
    per_page = IntegerField(
        required=True, max_value=DEFAULT_MAX_PAGE_SIZE,
        help_text=_('Page size'),
    )
    has_next = BooleanField(help_text=_('Has result next page'))
    url = CharField(help_text=_('Current page URL'))
    after_url = CharField(help_text=_('URL of page after cursor position'), allow_null=True)
    before_url = CharField(help_text=_('URL of page before cursor position'), allow_null=True)


class CursorUUIDResponseSerializer(CursorUUIDUncountedResponseSerializer):
    total = IntegerField(help_text=_('Total result size'))
    last_url = CharField(help_text=_('Last page URL'), allow_null=True)



from __future__ import annotations
import typing as t

GenericContext = t.Dict[str, t.Any]
ImmutableContext = t.Mapping[str, t.Any]

T = t.TypeVar('T')

SequenceOrT = t.Union[t.Sequence[T], T]

Prefix = str
Prefixes = t.Tuple[Prefix, ...]

MappingSet = t.Set[str]
OptionalMappingSet = t.Optional[MappingSet]

InitKwargs = t.Dict[str, t.Any]
OptionalInitKwargs = t.Optional[InitKwargs]



from __future__ import annotations
import copy
import typing
from rest_framework.routers import BaseRouter, SimpleRouter, Route

from restdoctor.rest_framework.resources import ResourceViewSet
from restdoctor.utils.constants import IDENTIFIER_REGEX, DEFAULT_MAPPING_SET

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.viewsets import ViewSetMixin

    from restdoctor.django.custom_types import URLPatternList
    from restdoctor.rest_framework.custom_types import RouteOrDynamicRouteList
    from restdoctor.utils.custom_types import OptionalMappingSet, OptionalInitKwargs

    MethodMap = typing.Dict[str, str]


class ResourceRouter(SimpleRouter):
    def get_method_map(self, viewset: ViewSetMixin, method_map: MethodMap) -> MethodMap:
        if not issubclass(viewset, ResourceViewSet):
            return super().get_method_map(viewset, method_map)

        actions = viewset.get_action_methods() + viewset.get_extra_actions()
        bound_methods = {}
        for method, action in method_map.items():
            if action in actions:
                bound_methods[method] = action
        return bound_methods


class TreeMixin(BaseRouter):
    def __init__(  # noqa: CFQ002
        self,
        parent_router: typing.Optional[BaseRouter] = None,
        parent_prefix: typing.Optional[str] = None,
        parent_viewset: typing.Any = None,
        mapping_set: OptionalMappingSet = None,
        initkwargs: OptionalInitKwargs = None,
        *args: typing.Any, **kwargs: typing.Any,
    ) -> None:
        self.nest_count = 0
        self.children: typing.List[TreeMixin] = []
        if parent_router:
            self.parent_router = parent_router
            self.nest_count = getattr(parent_router, 'nest_count', 0) + 1
            self.nest_prefix = kwargs.pop('lookup', 'nested_%i' % self.nest_count) + '_'

        super().__init__(*args, **kwargs)

        if not parent_router or not parent_prefix:
            return

        mapping_set = (mapping_set or DEFAULT_MAPPING_SET) & DEFAULT_MAPPING_SET

        if 'trailing_slash' not in kwargs:
            self.trailing_slash = parent_router.trailing_slash

        self.check_valid_name(self.nest_prefix)

        nested_routes: RouteOrDynamicRouteList = []
        parent_lookup_regex = parent_router.get_lookup_regex(parent_viewset, self.nest_prefix)

        self.parent_regex = f'{parent_prefix}/{parent_lookup_regex}/'

        self.routes: RouteOrDynamicRouteList
        for route in self.routes:
            initkwargs = copy.deepcopy(initkwargs or {})

            initkwargs.update(route.initkwargs)
            route_contents = route._asdict()
            escaped_parent_regex = self.parent_regex.replace('{', '{{').replace('}', '}}')
            route_contents['url'] = route.url.replace('^', '^' + escaped_parent_regex)
            route_contents['initkwargs'] = initkwargs
            # WIP: need the way to filter out some nested endpoints
            if isinstance(route, Route):
                mapping = {
                    method: view_function
                    for method, view_function in route.mapping.items()
                    if view_function in mapping_set
                }
                route_contents['mapping'] = mapping
            nested_routes.append(type(route)(**route_contents))

        self.routes = nested_routes

    def check_valid_name(self, value: str) -> None:
        if IDENTIFIER_REGEX.match(value) is None:
            raise ValueError(f"lookup argument '{value}' needs to be valid python identifier")

    def register_child(  # noqa: CFQ002
        self,
        prefix: str,
        viewset: typing.Any,
        basename: str,
        lookup: str,
        *args: typing.Any, **kwargs: typing.Any,
    ) -> TreeMixin:
        super().register(prefix, viewset, basename=basename)

        kwargs['parent_router'] = self
        kwargs['parent_prefix'] = prefix
        kwargs['parent_viewset'] = viewset
        kwargs['lookup'] = lookup
        child_router = self.__class__(*args, **kwargs)

        self.children.append(child_router)
        return child_router


class TreeRouter(TreeMixin, SimpleRouter):
    def get_urls(self) -> URLPatternList:
        urls = super().get_urls()
        for child in self.children:
            urls.extend(child.urls)

        return urls



from tests.test_unit.test_serializers.test_meta_inheritance.stubs import (
    MyModel, ParentModelSerializer, SerializerMixinA, SerializerMixinB,
)


def test_no_meta_in_child():
    class ChildSerializer(ParentModelSerializer):
        pass

    assert 'Meta' not in ChildSerializer.__dict__


def test_no_meta_model_in_parents():
    class ChildSerializer(SerializerMixinA, SerializerMixinB):
        class Meta:
            pass

    assert not hasattr(ChildSerializer.Meta, 'model')


def test_inherited_from_parent():
    class ChildSerializer(SerializerMixinA, ParentModelSerializer, SerializerMixinB):
        class Meta:
            pass

    assert 'Meta' in ChildSerializer.__dict__
    assert ChildSerializer.Meta.__bases__ == (object,)
    assert ChildSerializer.Meta.model is MyModel


def test_inherited_from_grandparent():
    class SubSerializer(ParentModelSerializer):
        pass

    class ChildSerializer(SubSerializer):
        class Meta:
            pass

    assert 'Meta' in ChildSerializer.__dict__
    assert ChildSerializer.Meta.__bases__ == (object,)
    assert ChildSerializer.Meta.model is MyModel


def test_is_not_inherited_if_defined_in_child():
    class MyChildModel(MyModel):
        class Meta:
            app_label = 'bestdoctor'

    class ChildSerializer(ParentModelSerializer):
        class Meta:
            model = MyChildModel

    assert ChildSerializer.Meta.model is MyChildModel



from __future__ import annotations

import typing
from urllib.parse import urljoin

from django.conf import settings
from rest_framework.generics import GenericAPIView
from rest_framework.request import Request
from rest_framework.schemas.openapi import SchemaGenerator

from restdoctor.rest_framework.schema.openapi import RestDoctorSchema
from restdoctor.rest_framework.schema.refs_registry import LocalRefsRegistry

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.custom_types import Handler
    from restdoctor.rest_framework.schema.custom_types import OpenAPISchema


class RefsSchemaGenerator(SchemaGenerator):
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)
        self.local_refs_registry = LocalRefsRegistry()

        self.api_version = settings.API_DEFAULT_VERSION
        urlconf = kwargs.get('urlconf')
        if urlconf:
            for api_version, api_urlconf in settings.API_VERSIONS.items():
                if api_urlconf == urlconf:
                    self.api_version = api_version
                    break
        self.api_default_format = settings.API_DEFAULT_FORMAT
        self.api_formats = settings.API_FORMATS

    def get_paths(self, request: Request = None) -> typing.Optional[OpenAPISchema]:
        result: OpenAPISchema = {}

        paths, view_endpoints = self._get_paths_and_endpoints(request)

        if not paths:
            return None

        for path, method, view in view_endpoints:
            if not self.has_view_permissions(path, method, view):
                continue
            operation = view.schema.get_operation(path, method)
            if path.startswith('/'):
                path = path[1:]
            path = urljoin(self.url or '/', path)

            result.setdefault(path, {})
            result[path][method.lower()] = operation

        return result

    def create_view(
        self, callback: Handler, method: str, request: Request = None
    ) -> GenericAPIView:
        view = super().create_view(callback, method, request)
        schema_class = getattr(view, 'schema_class', RestDoctorSchema)
        view.schema = schema_class(generator=self)
        return view

    def get_error_schema(
        self, description: str = 'Описание ошибки', detailed: bool = False
    ) -> OpenAPISchema:
        schema: OpenAPISchema = {
            'type': 'object',
            'properties': {'message': {'type': 'string', 'description': description}},
        }
        if detailed:
            schema['properties']['errors'] = {'type': 'array', 'items': {'type': 'object'}}
        return schema

    def get_schema(
        self, request: Request = None, public: bool = False
    ) -> typing.Optional[OpenAPISchema]:
        self._initialise_endpoints()
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/ErrorResponseSchemaDetailed', self.get_error_schema(detailed=True)
        )
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/ErrorResponseSchema', self.get_error_schema()
        )
        self.local_refs_registry.put_local_ref(
            '#/components/schemas/NotFoundResponseSchema',
            self.get_error_schema(description='Ресурс не найден.'),
        )

        paths = self.get_paths(None if public else request)
        if not paths:
            return None

        schema = {'openapi': '3.0.2', 'info': self.get_info(), 'paths': paths}
        components = self.local_refs_registry.get_components()
        if components:
            schema['components'] = components

        return schema



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from django.db.models import Model

from rest_framework.serializers import (
    ModelSerializer as BaseModelSerializer, Serializer as BaseSerializer,
    SerializerMetaclass as BaseSerializerMetaclass,
)


def _is_serializer_subclass(obj: Any) -> bool:
    try:
        return issubclass(obj, Serializer)
    except TypeError:
        return False


def _get_serializer_fields(serializer_cls: Serializer) -> List[str]:
    try:
        return list(serializer_cls.Meta.fields)
    except AttributeError:
        return list(serializer_cls._declared_fields)


class _DeferredMetaFields(tuple):
    pass


def extend_meta_fields(*args: Union[Serializer, str]) -> _DeferredMetaFields:
    return _DeferredMetaFields(args)


class SerializerMetaclass(BaseSerializerMetaclass):
    def __new__(cls, name: str, bases: Tuple, attrs: Dict[str, Any]) -> SerializerMetaclass:
        meta = attrs.get('Meta')
        if meta:
            fields = getattr(meta, 'fields', None)
            if isinstance(fields, _DeferredMetaFields):
                meta.fields = cls._extend_meta_fields(fields, bases)
        return super().__new__(cls, name, bases, attrs)

    @classmethod
    def _extend_meta_fields(cls, deferred_fields: _DeferredMetaFields, bases: Tuple) -> List[str]:
        fields: List[str] = []
        serializer_in_fields = False
        for arg in deferred_fields:
            if isinstance(arg, str):
                fields.append(arg)
            elif _is_serializer_subclass(arg):
                fields.extend(_get_serializer_fields(arg))
                serializer_in_fields = True
            else:
                raise TypeError(f'str or Serializer subclass expected, got {arg!r}')
        if serializer_in_fields:
            return fields
        parent_fields: List[str] = []
        for parent_cls in bases:
            parent_fields.extend(_get_serializer_fields(parent_cls))
        return parent_fields + fields


class ModelSerializerMetaclass(SerializerMetaclass):
    def __new__(cls, name: str, bases: Tuple, attrs: Dict[str, Any]) -> ModelSerializerMetaclass:
        serializer_cls = super().__new__(cls, name, bases, attrs)
        meta = attrs.get('Meta')
        if meta:
            if not hasattr(meta, 'model'):
                model = cls._get_meta_model(serializer_cls)
                if model:
                    meta.model = model
            fields = getattr(meta, 'fields', None)
            if isinstance(fields, _DeferredMetaFields):
                meta.fields = cls._extend_meta_fields(fields, bases)
        return serializer_cls

    @classmethod
    def _get_meta_model(cls, serializer_cls: ModelSerializer) -> Optional[Model]:
        for parent_cls in inspect.getmro(serializer_cls):
            parent_meta = parent_cls.__dict__.get('Meta')
            if parent_meta and hasattr(parent_meta, 'model'):
                return parent_meta.model
        return None


class Serializer(BaseSerializer, metaclass=SerializerMetaclass):
    pass


class EmptySerializer(Serializer):
    pass


class ModelSerializer(BaseModelSerializer, Serializer, metaclass=ModelSerializerMetaclass):
    pass



from typing import Optional

from setuptools import setup, find_packages


def get_version() -> Optional[str]:
    with open('restdoctor/__init__.py', 'r') as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith('__version__'):
            return line.split('=')[-1].strip().strip("'")


def get_long_description() -> str:
    with open('README.md', encoding='utf8') as f:
        return f.read()


setup(
    name='restdoctor',
    description='BestDoctor\'s batteries for REST services.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(),
    include_package_data=True,
    keywords='statistics',
    version=get_version(),
    author='BestDoctor',
    author_email='s.butkin@bestdoctor.ru',
    license='MIT',
    py_modules=['restdoctor'],
    zip_safe=False,
)



import pytest


@pytest.mark.parametrize(
    'method',
    [
        'delete',
        'get',
        'patch',
        'post',
        'put',
    ]
)
def test_call_api(method, drf_client, mock_api_call_factory):
    api_call = mock_api_call_factory(method)
    called_method = getattr(drf_client, method)

    called_method('test/path')

    api_call.assert_called_once_with(
        'test/path', HTTP_ACCEPT='application/json', HTTP_AUTHORIZATION='', content_type='application/json')


@pytest.mark.parametrize(
    ('method', 'content_type'),
    [
        ('delete', 'application/json'),
        ('get', 'multipart'),
        ('patch', 'multipart'),
        ('post', 'multipart'),
        ('put', 'multipart'),
    ]
)
def test_call_api_with_content_type(method, content_type, drf_client, mock_api_call_factory):
    api_call = mock_api_call_factory(method)
    called_method = getattr(drf_client, method)

    called_method('test/path', content_type=content_type)

    api_call.assert_called_once_with(
        'test/path', HTTP_ACCEPT='application/json', HTTP_AUTHORIZATION='', content_type=content_type)



from __future__ import annotations
import dataclasses
import typing

from django.conf import settings

from restdoctor.utils.api_prefix import get_api_prefix


@dataclasses.dataclass
class APIParams:
    accepted: str
    vendor: str
    prefix: typing.Optional[str] = None
    version: str = settings.API_FALLBACK_VERSION
    resource_discriminator: typing.Optional[str] = None
    format: str = settings.API_DEFAULT_FORMAT  # noqa: A003, VNE003

    @property
    def version_with_resource_discriminator(self) -> str:
        if self.resource_discriminator is None:
            return self.version
        else:
            return f'{self.version}-{self.resource_discriminator}'


def parse_accept(header: str = None, vendor: str = None) -> typing.Optional[APIParams]:
    if not header:
        return None
    api_params = APIParams(
        prefix=get_api_prefix(),
        accepted=header,
        vendor=vendor or 'vendor',
    )
    api_options_string = header.split('/', 1)[-1]
    if getattr(settings, 'API_FALLBACK_FOR_APPLICATION_JSON_ONLY', False):
        if api_options_string == 'json':
            return api_params
    elif not header.startswith(f'application/vnd.{api_params.vendor}'):
        return api_params

    api_params.version = settings.API_DEFAULT_VERSION

    api_options_string = api_options_string.split('+', 1)[0]

    parts = api_options_string.split('.')
    version = parse_version(parts)
    resource_discriminator = parse_resource_discriminator(parts)
    api_format = parse_api_format(parts)

    if version:
        api_params.version = version
    if resource_discriminator:
        api_params.resource_discriminator = resource_discriminator
    if api_format:
        api_params.format = api_format
    return api_params


def parse_version(api_options_parts: typing.List[str]) -> typing.Optional[str]:
    try:
        part = api_options_parts[2]
    except IndexError:
        pass
    else:
        version, *_ = part.split('-', 1)
        if version in settings.API_VERSIONS.keys():
            return version


def parse_resource_discriminator(api_options_parts: typing.List[str]) -> typing.Optional[str]:
    try:
        part = api_options_parts[2]
        return part.split('-', 1)[1]
    except IndexError:
        pass


def parse_api_format(api_options_parts: typing.List[str]) -> typing.Optional[str]:
    try:
        api_format = api_options_parts[3]
    except IndexError:
        pass
    else:
        if api_format in settings.API_FORMATS:
            return api_format


def get_api_header(params: APIParams) -> str:
    return f'{params.vendor}.{params.version_with_resource_discriminator}; format={params.format}'


def get_media_type(params: APIParams) -> str:
    if params.accepted.startswith('application/json'):
        return params.accepted

    return f'application/vnd.{params.vendor}.{params.version_with_resource_discriminator}.{params.format}+json'



from __future__ import annotations
import typing as t

from django.http import HttpRequest, HttpResponse
from django.urls.resolvers import URLPattern

URLPatternList = t.List[URLPattern]
DjangoHandler = t.Callable[[HttpRequest], HttpResponse]



import pytest

from restdoctor.utils.permissions import get_permission_classes_from_map
from tests.test_unit.stubs import (
    permission_classes_map_with_default, permission_classes_map_no_default,
    PermissionC, PermissionB, PermissionA,
)

@pytest.mark.parametrize(
    'permission_classes_map,action,default,expected',
    (
        (permission_classes_map_with_default, 'retrieve', [PermissionC], [PermissionB]),
        (permission_classes_map_with_default, 'list', [PermissionC], [PermissionA]),
        (permission_classes_map_no_default, 'retrieve', [PermissionC], [PermissionB]),
        (permission_classes_map_no_default, 'list', [PermissionC], [PermissionC]),
    ),
)
def test_get_permission_classes_from_map(permission_classes_map, action, default, expected):
    permission_classes = get_permission_classes_from_map(action, permission_classes_map, default)

    assert permission_classes == expected



from __future__ import annotations

import typing

from django.conf import settings

from restdoctor.rest_framework.schema.openapi import RestDoctorSchema
from restdoctor.rest_framework.schema.utils import get_action

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.custom_types import Handler, ResourceHandlersMap
    from restdoctor.rest_framework.generics import GenericAPIView
    from restdoctor.rest_framework.schema.custom_types import CodeActionSchemaTuple, OpenAPISchema


def get_single_or_default_handler(view: GenericAPIView) -> typing.Optional[Handler]:
    keys = list(view.resource_handlers_map.keys())
    if len(keys) == 1:
        return view.resource_handlers_map[keys[0]]
    return view.resource_handlers_map.get(view.default_discriminative_value)


class ResourceSchema(RestDoctorSchema):
    def get_object_name(self, path: str, method: str, action_name: str) -> str:
        return self.get_object_name_by_view_class_name(
            clean_suffixes=['View', 'APIView', 'ViewSet']
        )

    def get_action_code_schemas(
        self, path: str, method: str
    ) -> typing.Iterator[CodeActionSchemaTuple]:
        codes_seen: typing.Set[str] = set()

        if self.generator:
            for _, handler in self.get_resources(method).items():
                view = self.generator.create_view(handler, method, request=self.view.request)
                for code, serializer, schema in view.schema.get_action_code_schemas(path, method):
                    if code not in codes_seen:
                        yield code, serializer, schema

    def get_resources(self, method: str) -> ResourceHandlersMap:
        return {
            resource: handler
            for resource, handler in self.view.resource_handlers_map.items()
            if (
                method in self.view.resource_discriminate_methods
                or resource == self.view.default_discriminative_value
            )
        }

    def get_resources_request_body_schema(self, path: str, method: str) -> OpenAPISchema:
        schemas = {}
        if self.generator:
            for resource, handler in self.get_resources(method).items():
                view = self.generator.create_view(handler, method, request=self.view.request)
                if hasattr(view, get_action(path, method, view)):
                    schemas[resource] = view.schema.get_request_body_schema(path, method)

        list_schemas = []
        for schema in schemas.values():
            if schema not in list_schemas:
                list_schemas.append(schema)
        if len(list_schemas) == 1:
            return list_schemas[0]
        return {'oneOf': list_schemas}

    def get_resources_response_schema(
        self, path: str, method: str, api_format: str = None
    ) -> OpenAPISchema:
        schemas = {}
        if self.generator:
            for resource, handler in self.get_resources(method).items():
                view = self.generator.create_view(handler, method, request=self.view.request)
                if hasattr(view, get_action(path, method, view)):
                    schemas[resource] = view.schema.get_response_schema(
                        path, method, api_format=api_format
                    )

        list_schemas = []
        for schema in schemas.values():
            if schema not in list_schemas:
                list_schemas.append(schema)
        if len(list_schemas) == 1:
            return list_schemas[0]
        return {'oneOf': list_schemas}

    def get_content_schema_by_type(self, path: str, method: str, schema_type: str) -> OpenAPISchema:
        content_schema = {}

        if schema_type == 'responses':
            resources_schema_method_name = 'get_resources_response_schema'
            schema_method_name = 'get_response_schema'
        else:
            resources_schema_method_name = 'get_resources_request_body_schema'
            schema_method_name = 'get_request_body_schema'

        vendor = getattr(settings, 'API_VENDOR_STRING', 'vendor').lower()
        default_content_type = f'application/vnd.{vendor}'

        resources_schema_method = getattr(self, resources_schema_method_name)
        content_schema[default_content_type] = {'schema': resources_schema_method(path, method)}

        if not self.generator:
            return content_schema

        version_content_type = f'{default_content_type}.{self.generator.api_version}'

        for resource, handler in self.view.resource_handlers_map.items():
            view = self.generator.create_view(handler, method, request=self.view.request)
            if not hasattr(view, get_action(path, method, view)):
                continue
            schema_method = getattr(view.schema, schema_method_name)
            default_resource_schema = schema_method(path, method)

            resource_content_type = f'{version_content_type}-{resource}'

            content_schema[resource_content_type] = {'schema': default_resource_schema}

            for api_format in self.generator.api_formats:
                if api_format != self.generator.api_default_format:
                    resource_schema = schema_method(path, method, api_format=api_format)
                    if resource_schema != default_resource_schema:
                        content_type = f'{resource_content_type}.{api_format}'
                        content_schema[content_type] = {'schema': resource_schema}

        return content_schema

    def get_request_body(self, path: str, method: str) -> OpenAPISchema:
        if method not in ('PUT', 'PATCH', 'POST'):
            return {}

        return {'content': self.get_content_schema_by_type(path, method, 'request_body')}

    def get_pagination_parameters(self, path: str, method: str) -> typing.List[OpenAPISchema]:
        handler = get_single_or_default_handler(self.view)

        if self.generator and handler:
            view = self.generator.create_view(handler, method, request=self.view.request)
            return view.schema.get_pagination_parameters(path, method)

        return []

    def get_filter_parameters(self, path: str, method: str) -> typing.List[OpenAPISchema]:
        handler = get_single_or_default_handler(self.view)

        if self.generator and handler:
            view = self.generator.create_view(handler, method, request=self.view.request)
            return view.schema.get_filter_parameters(path, method)

        return []

    def get_request_serializer_filter_parameters(
        self, path: str, method: str
    ) -> typing.List[OpenAPISchema]:
        handler = get_single_or_default_handler(self.view)

        if self.generator and handler:
            view = self.generator.create_view(handler, method, request=self.view.request)
            return view.schema.get_request_serializer_filter_parameters(path, method)

        return []

    def _get_resources(self, method: str) -> ResourceHandlersMap:
        return self.get_resources(method)

    def __get_item_schema(
        self, path: str, method: str, api_format: str = None
    ) -> typing.Optional[OpenAPISchema]:
        schemas = {}
        if self.generator:
            for resource, handler in self._get_resources(method).items():
                view = self.generator.create_view(handler, method, request=self.view.request)
                schemas[resource] = view.schema._get_item_schema(
                    path, method, api_format=api_format
                )

        list_schemas = list(schemas.values())
        if len(list_schemas) == 1:
            return list_schemas[0]
        return {'oneOf': list_schemas}



from __future__ import annotations

from rest_framework.exceptions import APIException
from rest_framework import status

from restdoctor.constants import HTTP_420_GO_TO_HELL


class BadRequest(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = 'Есть ошибки в формировании запроса.'
    default_code = 'bad_request'


class BusinessLogicException(APIException):
    status_code = HTTP_420_GO_TO_HELL
    default_detail = 'Что-то пошло не так, обратитесь в поддержку.'
    default_code = 'business_logic_error'



from __future__ import annotations

from django.db import models
from rest_framework.fields import CharField, SerializerMethodField
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer, Serializer

from restdoctor.rest_framework.generics import GenericAPIView
from restdoctor.rest_framework.resources import ResourceViewSet
from restdoctor.rest_framework.schema import SchemaWrapper
from restdoctor.rest_framework.serializers import ModelSerializer
from restdoctor.rest_framework.views import RetrieveAPIView
from restdoctor.rest_framework.viewsets import ListModelViewSet, ModelViewSet, ReadOnlyModelViewSet


class ModelWithoutSensitiveData(models.Model):
    class Meta:
        app_label = 'restdoctor'

    title = models.CharField(max_length=100, blank=True)


class ChildSensitiveDataModel(models.Model):
    class Meta:
        app_label = 'restdoctor'

    class SensitiveData:
        include = ['title']

    title = models.CharField(max_length=100, blank=True)
    field_fk = models.ForeignKey(
        ModelWithoutSensitiveData, on_delete=models.PROTECT, blank=True, null=True
    )


class ParentSensitiveDataModel(models.Model):
    class Meta:
        app_label = 'restdoctor'

    class SensitiveData:
        include = ['title']

    title = models.CharField(max_length=100, blank=True)
    last_name = models.CharField(max_length=100, blank=True)
    field_fk = models.ForeignKey(
        ChildSensitiveDataModel, on_delete=models.PROTECT, blank=True, null=True
    )
    field_m2m = models.ManyToManyField(ChildSensitiveDataModel, blank=True)
    field_o2o = models.OneToOneField(
        ChildSensitiveDataModel, on_delete=models.PROTECT, blank=True, null=True
    )


class SerializerWithSensitiveData(Serializer):
    class SensitiveData:
        include = ['field1', 'field2']

    field1 = CharField()
    field2 = CharField()


class SerializerWithoutSensitiveData(Serializer):
    field1 = CharField()


class ModelSerializerWithSensitiveData(ModelSerializer):
    class Meta:
        model = ParentSensitiveDataModel
        fields = [
            'first_name',
            'last_name',
            'title',
            'field_fk',
            'field_m2m',
            'field_o2o',
            'field1',
            'field2',
            'field3',
        ]

    class SensitiveData:
        include = ['first_name', 'last_name']

    first_name = CharField()
    field1 = SerializerWithSensitiveData()
    field2 = SchemaWrapper(SerializerMethodField(), schema_type=SerializerWithSensitiveData)
    field3 = SerializerWithSensitiveData(many=True)


class PermissionA(BasePermission):
    pass


class PermissionB(BasePermission):
    pass


class PermissionC(BasePermission):
    pass


permission_classes_map_with_default = {'default': [PermissionA], 'retrieve': [PermissionB]}


permission_classes_map_no_default = {'retrieve': [PermissionB]}


class ModelA(models.Model):
    class Meta:
        app_label = 'restdoctor'


class ModelB(models.Model):
    class Meta:
        app_label = 'restdoctor'


class ModelAMixin:
    queryset = ModelA.objects.none()


class ModelBMixin:
    queryset = ModelB.objects.none()


class ModelAViewSet(ModelViewSet):
    queryset = ModelA.objects.none()


class ModelAView(RetrieveAPIView):
    queryset = ModelA.objects.none()


class ModelAWithMixinViewSet(ModelAMixin, ModelViewSet):
    pass


class ModelBViewSet(ModelViewSet):
    queryset = ModelB.objects.none()


class ModelBWithMixinViewSet(ModelBMixin, ModelViewSet):
    pass


class NoneViewSet(ModelViewSet):
    pass


class SerializerA(BaseSerializer):
    pass


class SerializerB(BaseSerializer):
    pass


class ListViewSetWithRequestSerializer(ListModelViewSet):
    serializer_class_map = {'default': SerializerA, 'list': {'request': SerializerB}}


class ListViewSetWithoutRequestSerializer(ListModelViewSet):
    serializer_class_map = {'default': SerializerA}


class ROViewSet(ReadOnlyModelViewSet):
    serializer_class_map = {'default': SerializerA}

    def dispatch(self, *args, **kwargs):
        return super(GenericAPIView).dispatch(*args, **kwargs)


class RWViewSet(ModelViewSet):
    serializer_class_map = {'default': SerializerA}

    def dispatch(self, *args, **kwargs):
        return super(GenericAPIView, self).dispatch(*args, **kwargs)

    def update(self, *args, **kwargs):
        return Response({})


class ComplexResourceViewSet(ResourceViewSet):
    resource_views_map = {'read_only': ROViewSet, 'read_write': RWViewSet}



import pytest
from django.urls import resolve
from rest_framework.routers import SimpleRouter

from restdoctor.rest_framework.schema import RefsSchemaGenerator


class UrlConf:
    urlpatterns = []


@pytest.fixture()
def get_create_view_func():
    def with_args(prefix, viewset, basename, router=None):
        router = router or SimpleRouter()
        generator = RefsSchemaGenerator()

        router.register(prefix, viewset, basename=basename)
        urlconf = UrlConf()
        urlconf.urlpatterns = router.urls

        def create_view(path, method):
            match = resolve(path, urlconf)
            view = generator.create_view(match.func, method, None)
            return view

        return create_view
    return with_args


@pytest.fixture()
def get_resource_ref():
    def with_attrs(name, suffix):
        return {'$ref': f'#/components/schemas/{name}{suffix}'}
    return with_attrs


@pytest.fixture()
def resource_default_ref(get_resource_ref):
    def with_attrs(suffix):
        return get_resource_ref('tests_Default', suffix)
    return with_attrs


@pytest.fixture()
def resource_another_ref(get_resource_ref):
    def with_attrs(suffix):
        return get_resource_ref('tests_Another', suffix)
    return with_attrs


@pytest.fixture()
def get_object_schema():
    def with_attrs(schema):
        return {'type': 'object', 'properties': {'data': schema}}
    return with_attrs


@pytest.fixture()
def resource_default_rq_schema(get_object_schema, resource_default_ref):
    return get_object_schema(resource_default_ref('RQ'))


@pytest.fixture()
def resource_default_wq_schema(get_object_schema, resource_default_ref):
    return get_object_schema(resource_default_ref('WQ'))


@pytest.fixture()
def resource_another_rq_schema(get_object_schema, resource_another_ref):
    return get_object_schema(resource_another_ref('RQ'))


@pytest.fixture()
def resource_another_wq_schema(get_object_schema, resource_another_ref):
    return get_object_schema(resource_another_ref('WQ'))
